########################################################################
# Copyright (C) 1988,1989,1990
# FourGen Software Technologies, Inc. Edmonds, Washington USA
# Use, modification, duplication, and/or distribution of this
# software is limited to the terms of the software agreement.
# Sccsid:  @(#)  .../data/all/triggers.4gl  1.5  Delta: 9/20/90
########################################################################

globals "globals.4gl"

define
    l_where char(200),
    l_trigger char(20),
    l_scratch char(400)

#####################################################################
function trigger(trig_text)
# returning true/false
#####################################################################
#
# Available triggers:
#
# before init
#   - called after defaults are set & before the main program logic
#   - set global variables
#   - change assigned defaults
# before schema
#   - called once per table before any schema work is done on that table
# before load
#   - called once per table before data is loaded into that table
#     from the .unl file (if the -l flag was passed on the command line)
#   - used primarily to set the del_filter variable
#     that defines the rows to delete (if any) before loading data.
#     (default: "1=0" - no delete)
# before unload
#   - called once per table before data is unloaded from that table
#     to the .unl file (if the -u flag was passed on the command line)
#   - used primarily to set the unl_filter variable
#     that defines the rows (from this table) to unload.
#     (default: "1=1" - unload all rows)
#     and to set the data unload order_by variable
#        (default: null - no order)
#        (sample order_by string: "doc_no desc, doc_date")
# before error
#   - called after an sql error has occured, and before the error
#    routines are run.
#  - primarily used to intelligently set the g_err_action variable
#    to either "continue", "quit", or "ask" so the error routines
#    can be invoked accordingly.
#
    define
        trig_text char(20),
        c char(1),             # input character
        n smallint             # working number

    # blank out the where clause & copy trig_text to static variable
    let l_where = null
    let l_trigger = trig_text

    case
      when trig_text = "before all indexes"
        # 3.11 had a bad index name.  this will drop it.
        select idxname from sysindexes, systables
          where
            sysindexes.idxname = "i1xaddlr" and
            sysindexes.tabid = systables.tabid and
            systables.tabname = "stxaddld"
        if sqlca.sqlcode = 0
        then
            drop index i1xaddlr
        end if
      when tab[gc_tab].tabname = "pcdtablr"
        let order_by = "prodid, line_no"
        let l_where = "prodid = ",g_quote,"all",g_quote
        return t_system()

      when tab[gc_tab].tabname = "stxprogr"
        let order_by = "module, progname"
        let l_where  = "module = ",g_quote,"all",g_quote
        return t_system()

      when tab[gc_tab].tabname = "stxcompr"
        let order_by = "comp_id"
        return t_initialize()

      when tab[gc_tab].tabname = "stxcntrc"
        return t_control()

      when tab[gc_tab].tabname = "stxperdr"
        if trig_text = "before load"
        then
            # when updating the perdr table to 3.0, the balanced
            # column needs to be set to "Y" (under g/l control)
            update stxperdr set balanced="Y" where balanced is null
        end if
        let order_by = "period_year, period"
        return t_regular()

      when tab[gc_tab].tabname = "stxchrtr"
        let order_by = "acct_no"
        return t_regular()

      when tab[gc_tab].tabname = "stxchrtd"
        # loading and unloading is handled by the G/L package
        if trig_text = "before load" then return false end if
        if trig_text = "before unload" then return false end if
        return true

      when tab[gc_tab].tabname = "stxtranr"
        # loading and unloading is handled by the other ledgers
        if trig_text = "before load" then return false end if
        if trig_text = "before unload" then return false end if
        return true

      when tab[gc_tab].tabname = "stxinfor"
        let order_by = "src_type, src_key"
        let l_where = "src_type = ",g_quote,"D",g_quote,
              " or src_type = ",g_quote,"A",g_quote,
              " or src_type = ",g_quote,"S",g_quote
        return t_initialize()

      when tab[gc_tab].tabname = "stxactgr"
        let order_by = "grp_key"
        return t_initialize()

      when tab[gc_tab].tabname = "stxactgd"
        let order_by = "grp_key, line_no"
        return t_initialize()

      when tab[gc_tab].tabname = "stxerorh"
        let order_by = "err_module, err_program, err_number"
        let l_where  = "err_module = ",g_quote,"all",g_quote,
              " or err_module = ",g_quote,"lib_all",g_quote,
              " or err_module = ",g_quote,"r_menu",g_quote
        return t_system()

      when tab[gc_tab].tabname = "stxerord"
        let order_by = "err_module, err_program, err_number,a_b,line_no"
        let l_where  = "err_module = ",g_quote,"all",g_quote,
              " or err_module = ",g_quote,"lib_all",g_quote,
              " or err_module = ",g_quote,"r_menu",g_quote
        return t_system()

      when tab[gc_tab].tabname = "stxhelph"
        let order_by = "hlp_module, hlp_program, hlp_number"
        let l_where  = "hlp_module = ",g_quote,"all",g_quote,
          " or hlp_module = ",g_quote,"lib",g_quote,
          " or hlp_module = ",g_quote,"lib_all",g_quote,
          " or hlp_module = ",g_quote,"r_menu",g_quote
        return t_system()

      when tab[gc_tab].tabname = "stxhelpd"
        let order_by = "hlp_module, hlp_program, hlp_number, line_no"
        let l_where  = "hlp_module = ",g_quote,"all",g_quote,
          " or hlp_module = ",g_quote,"lib",g_quote,
          " or hlp_module = ",g_quote,"lib_all",g_quote,
          " or hlp_module = ",g_quote,"r_menu",g_quote
        return t_system()

      when tab[gc_tab].tabname = "stxaddlr"
        let order_by = "filename, line_no"
        return t_regular()

      when tab[gc_tab].tabname = "stxaddld"
        let order_by = "filename, record_key, line_no"
        return t_regular()

      when tab[gc_tab].tabname = "stxckrgr"
        let order_by = "acct_no, department"
        return t_activity()

      when tab[gc_tab].tabname = "stxckrgd"
        let order_by = "acct_no, department, orig_journal, doc_no"
        return t_activity()

      when tab[gc_tab].tabname = "stgtranr"
        let order_by = "orig_journal, doc_no"
        return t_activity()

      when tab[gc_tab].tabname = "stgactvd"
        let order_by = "orig_journal, doc_no, acct_no, department"
        return t_activity()

      when tab[gc_tab].tabname = "stxfiltr"
        if trig_text = "before unload" then return false end if
        if trig_text = "before load" then return false end if
        return true

    end case
    return true
end function
# trigger()

##  The following t_* functions are functions that work based on
##  the type of table you have:
##
##  t_regular()
##    On data load:
##      - if database is sample, then reload (delete, then load)
##      - if database isn't sample, then load on top of existing data
##    On data unload:
##      - if database isn't sample, then error if any data in table
##        (normally, you don't want data to be unloaded from this table)
##
##  t_activity()
##    Same as t_regular, only it warns you if there is any data
##    in ANY database (even sample).
##
##  t_initialize()
##    On data load:
##      - if database is sample, then reload (delete, then load)
##      - if database isn't sample, then load ONLY if no data exists
##    On data unload:
##      - set unl_filter to l_where if l_where isn't null
##        (nothing special occurs for unload)
##
##  t_control()
##    Same as t_initialize(), only it checks for exactly 1 row during
##    load and unload.
##
##  t_system()
##    This is for system maintained tables.  It always overwrites
##    (deletes first, then loads) tables regardless of the database.
##

#####################################################################
function t_control()
# returning true/false
#####################################################################
#  This function is for control tables.  It returns false if you
#  should not load data into the control table (data exists)
#
    return t_init(true)
end function
# t_control()

#####################################################################
function t_initialize()
# returning true/false
#####################################################################
#  This function is for tables that need data loaded only once.
#  like control tables, only no checking for 1 row.
#
    return t_init(false)
end function
# t_initialize()

#####################################################################
function t_system()
# returning true
#####################################################################
#  This function is for system maintained tables.
#  It always REloads the data (sample and non-sample databases)
#
    case
      when l_trigger = "before unload"
        if l_where is not null then let unl_filter = l_where end if
      when l_trigger = "before load"
        if l_where is null
          then let del_filter = "1=1"
          else let del_filter = l_where
        end if
    end case
    return true
end function
# t_system()

#####################################################################
function t_regular()
# returning true/false
#####################################################################
#  This function is for regular (non system, non control) type tables.
#  On unload, it errors if there is any activity in a non-sample
#  database to be unloaded.
#  On load, it always reloads (delete first) if the database is sample,
#  and always overloads (no delete first) if the database is not sample.
#  (Normally, there is no data to load into the tables if the database
#  isn't sample, unless you are transferring data from one accounting
#  system to another into a new database)
#
    return t_work(false)
end function
# t_regular()

#####################################################################
function t_activity()
# returning true/false
#####################################################################
#  This function is called for activity (posted) type tables.
#  It differs from t_regular in that it warns you if there is
#  any activity to unload in the sample database
#  (we usually don't unload posted data from sample or non-sample).
#
    return t_work(true)
end function
# t_activity()

###########
# The following functions are called by the other t_* functions

#####################################################################
function t_work(p_warn)
# returning true/false
#####################################################################
#  This function is the guts of t_activity() and t_regular().
#
    define
        p_warn smallint,  # boolean: warn if data is in the sample db?
        n smallint

    case
      when l_trigger = "before init" or l_trigger = "before schema"
        return true
      when l_trigger = "before unload"
        if l_where is not null then let unl_filter = l_where end if
        if p_warn = false and g_database = "sample"
          then return true
        end if

        # See if there is any data in the table & warn the operator.
        let l_scratch = "select count(*) from ",tab[gc_tab].tabname
        if l_where is not null
        then
            let l_scratch = l_scratch clipped, " where ", l_where
        end if
        prepare get_work from l_scratch
        declare work_curs cursor for get_work
        open work_curs
        fetch work_curs into n
        close work_curs
        if n > 0
        then
            let l_scratch = "DATA detected in ",
              tab[gc_tab].tabname
            call err_log(l_scratch)
        end if
      when l_trigger = "before load"
        if g_database = "sample"
        then
            if l_where is null
              then let del_filter = "1=1"   # replace everything
              else let del_filter = l_where # replace selected data
            end if
        else
            let del_filter = "1=0"   # don't replace-load over
        end if
    end case
    return true
end function
# t_work()

#####################################################################
function t_init(p_control)
# returning true/false
#####################################################################
#  This function is used for initializing data in tables.
#  If data exists in a non-sample database, then it doesn't reload.
#  It always reloads if the database is sample.
#  If p_control is passed as true, then it also checks for exactly
#  one row in the table.
#
    define
        p_control smallint, # boolean: check for 1 row (control table)
        n smallint # working number

    if l_trigger = "before init" or l_trigger = "before schema"
      then return true
    end if

    # prepare the select statement
    let l_scratch = "select count(*) from ",tab[gc_tab].tabname
    if l_where is not null
      then let l_scratch = l_scratch clipped, " where ", l_where
    end if
    prepare get_cont from l_scratch
    declare cont_curs cursor for get_cont
    open cont_curs

    case
      when l_trigger = "before unload"
        if l_where is not null then let unl_filter = l_where end if

        # no need to check for 1 row if not a control type table
        if p_control = false then return true end if
        fetch cont_curs into n
        close cont_curs
        if n != 1
        then
            let l_scratch = "UNLOAD: Not exactly 1 row in ",
              tab[gc_tab].tabname
            call err_log(l_scratch)
        end if

      when l_trigger = "before load"
        if g_database = "sample"
        then
            if l_where is null
              then let del_filter = "1=1"     # replace with the
              else let del_filter = l_where   # new sample data
            end if
            close cont_curs
            return true
        end if
        fetch cont_curs into n
        close cont_curs
        case
          when n = 0 return true  # ok to load the row
          otherwise
            if n > 1 and p_control
            then
                let l_scratch = "More than 1 row in ",
                  tab[gc_tab].tabname
                call err_log(l_scratch)
            end if
            return false
        end case
    end case
    return true
end function
# t_init()
